import 'dart:io';

import 'package:analyzer/analyzer.dart';
import 'package:analyzer/dart/analysis/features.dart';
import 'package:analyzer/dart/ast/standard_ast_factory.dart';
import 'package:analyzer/dart/analysis/utilities.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/ast/token.dart';
import 'package:analyzer/error/error.dart';
import 'package:analyzer/src/dart/scanner/scanner.dart';
import 'package:analyzer/src/generated/parser.dart';
import 'package:analyzer/src/string_source.dart';
import 'package:analyzer/dart/ast/visitor.dart';
import 'package:_fe_analyzer_shared/src/scanner/token.dart';
import 'package:dart_style/dart_style.dart';
import 'common/common.dart';
import 'package:args/args.dart';
import 'dart:convert';


void main(List<String> args) {
  final parser = ArgParser();
  parser.addOption('dir', defaultsTo: '');

  final results = parser.parse(args);
  final directoryPath = results['dir'];
  addShowcasePackageIfMissing(path: directoryPath);
  autoGenerateProject(path: directoryPath);
}

//Todo: Change this. dev_dependencies could be before  dependencies
void addShowcasePackageIfMissing({String path = ''}) {
  final pubspec = File('${path}pubspec.yaml');
  final lines = pubspec.readAsLinesSync();

  var dependencyLine;
   for (final line in lines.asMap().entries) {
     if(line.value.startsWith('dependencies:')) {
       dependencyLine =  line.key;
     } else  {
       if(line.value.contains('flutter_showcase:')) {
         dependencyLine = null;
         return;
       }
     }
   }

  if(dependencyLine != null) {
    lines.insert(dependencyLine +1, '  flutter_showcase:');
    final newFile = lines.join('\n');
    pubspec.writeAsStringSync(newFile);
  }

}

void autoGenerateProject({String path = ''}) {
  final featureSet = FeatureSet.fromEnableFlags([
    'extension-methods',
    'non-nullable',
  ]);

  final filePath = '${path}lib/main.dart';
  final fileContent = File(filePath).readAsStringSync();

  // TODO: Add support for  WidgetApp and CupertinoApp
  if (fileContent.contains('MaterialApp') && fileContent.contains('main')) {
    final parseResult = parseString(
      content: fileContent,
      featureSet: featureSet,
      path: filePath,
      throwIfDiagnostics: false,
    );

    final unit = parseResult.unit;

    importPackageIfMissing(unit);

    var v = AddShowcaseVisitor();
    unit.visitChildren(v);

    var formatter = new DartFormatter();

    try {
      final newContent = formatter.format(unit.toSource());
      File(filePath).writeAsStringSync(newContent);
    } on FormatterException catch (ex) {
      print(ex);
    }
    print('Adding showcase to project succesfully');
  } else {
    throwToolExit('There is no MaterialApp in lib/main.dart');
  }
}


void importPackageIfMissing (CompilationUnit unit) {
  final packageImport = 'import \'package:flutter_showcase/flutter_showcase.dart\';';
  if (!unit.toSource().contains(packageImport)) {
    final Directive node = unit.directives.last;

    final Directive clone = AstCloner().cloneNode(node);

    int offset = clone.offset;
    StringToken importToken = StringToken(TokenType.STRING, 'import', offset);
    StringToken semiColonToken = StringToken(TokenType.STRING, ';', offset + packageImport.length -1);

    final packageString = '\'package:flutter_showcase/flutter_showcase.dart\'';
    StringToken st = StringToken(TokenType.STRING, packageString, offset + importToken.length);
    final ssl = astFactory.simpleStringLiteral(st, packageString);
    final ImportDirective newDirectory = astFactory.importDirective(
        null,
        null,
        importToken,
        ssl,
        null,
        null,
        null,
        null,
        null,
        semiColonToken);


    clone.childEntities.forEach((element) {
      print(element);
    });
    /*  String newString = node.toSource() + packageImport;
      int offset = node.offset;
      StringToken st = StringToken(TokenType.STRING, newString, offset);
      final ssl = astFactory.simpleStringLiteral(st, newString);
      final NodeReplacer replacer = NodeReplacer(node, ssl);*/

    unit.directives.add(newDirectory);
  }

}

class AddShowcaseVisitor extends RecursiveAstVisitor {


  @override
  visitMethodInvocation(MethodInvocation node) {
    //Check if Material App exits
    if (node.toSource().contains('MaterialApp')) {
      // Check if [builder] argument is called
      if (node.toSource().contains('builder:')) {
        final builderNode = node.argumentList.arguments
            .where((element) => element.toSource().contains('builder:'))
            .first;
        // Check if [builder] argument contains already a frame  builder
        if (!builderNode.toSource().contains('Frame.builder') ||
            !builderNode.toSource().contains('FrameBuilder')) {
          print(builderNode.toSource());
          String newString =
              'builder: (context, child) => FrameBuilder(app: child, ${builderNode
              .toString()})';
          int offset = builderNode.offset;
          StringToken st = StringToken(TokenType.STRING, newString, offset);

          final ssl = astFactory.simpleStringLiteral(st, newString);
          final NodeReplacer replacer = NodeReplacer(builderNode, ssl);
          builderNode.parent.accept(replacer);
        }
      } else {
        // Add frame builder
        final builderNode = node.argumentList.arguments.first;
        String newString = ' ${builderNode.toString()}, builder: Frame.builder';
        int offset = builderNode.offset;
        StringToken st = StringToken(TokenType.STRING, newString, offset);
        final ssl = astFactory.simpleStringLiteral(st, newString);
        final NodeReplacer replacer = NodeReplacer(builderNode, ssl);
        builderNode.parent.accept(replacer);
      }
    }
    if (node.toSource().contains('runApp') && !node.toSource().contains('Showcase')) {
      final widget = node.argumentList.arguments.first;

      SimpleStringLiteral ssl = _createSimpleStringLiteral(widget);
      final NodeReplacer replacer = NodeReplacer(widget, ssl);
      widget.parent.accept(replacer);
    }
  }
}

SimpleStringLiteral _createSimpleStringLiteral(AstNode node) {
  String newString = modify(node.toString());
  int offset = node.offset;
  StringToken st = StringToken(TokenType.STRING, newString, offset);

  return astFactory.simpleStringLiteral(st, newString);
}

String modify(String s) {
  return 'Showcase(app: $s)';
}
/*


SimpleStringLiteral _create_SimpleStringLiteral(AstNode node){
  String new_string = modify(node.toString());
  int line_num = node.offset;
  //holds the position and type
  StringToken st = new StringToken(
      TokenType.STRING,new_string,line_num);
  return new SimpleStringLiteral(st, new_string);
}
String modify(String s){
  List parts = s.split('=');
  var value = parts[1];
  List l = parts[0].split('.');
  String dynamism = l.sublist(0,l.length-1).join('.');
  String propertyName = l.last.trim();
  return '${dynamism}.set("${propertyName}",${value})';
}*/
